@@ -29,7 +29,13 @@ from builtins import isinstance
 from typing import Iterable
 
 from .connection import ConnectionManager
-from .exceptions import KeycloakGetError, KeycloakPostError, raise_error_from_response
+from .exceptions import (
+    KeycloakDeleteError,
+    KeycloakGetError,
+    KeycloakPostError,
+    KeycloakPutError,
+    raise_error_from_response,
+)
 from .keycloak_openid import KeycloakOpenID
 from .urls_patterns import (
     URL_ADMIN_AUTHENTICATOR_CONFIG,
@@ -417,7 +423,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": realm_name}
         data_raw = self.raw_put(URL_ADMIN_REALM.format(**params_path), data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def delete_realm(self, realm_name):
         """
@@ -429,7 +435,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": realm_name}
         data_raw = self.raw_delete(URL_ADMIN_REALM.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_users(self, query=None):
         """
@@ -500,7 +506,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "alias": idp_alias}
         data_raw = self.raw_delete(URL_ADMIN_IDP.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def create_user(self, payload, exist_ok=False):
         """
@@ -592,7 +598,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": user_id}
         data_raw = self.raw_put(URL_ADMIN_USER.format(**params_path), data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def delete_user(self, user_id):
         """
@@ -604,7 +610,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": user_id}
         data_raw = self.raw_delete(URL_ADMIN_USER.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def set_user_password(self, user_id, password, temporary=True):
         """
@@ -625,7 +631,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_RESET_PASSWORD.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def get_credentials(self, user_id):
         """
@@ -658,7 +664,7 @@ class KeycloakAdmin:
             "credential_id": credential_id,
         }
         data_raw = self.raw_delete(URL_ADMIN_USER_CREDENTIAL.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakDeleteError)
 
     def user_logout(self, user_id):
         """
@@ -715,7 +721,7 @@ class KeycloakAdmin:
         data_raw = self.raw_post(
             URL_ADMIN_USER_FEDERATED_IDENTITY.format(**params_path), data=json.dumps(payload)
         )
-        raise_error_from_response(data_raw, KeycloakPostError, expected_codes=[201])
+        return raise_error_from_response(data_raw, KeycloakPostError, expected_codes=[201, 204])
 
     def delete_user_social_login(self, user_id, provider_id):
 
@@ -727,7 +733,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": user_id, "provider": provider_id}
         data_raw = self.raw_delete(URL_ADMIN_USER_FEDERATED_IDENTITY.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def send_update_account(
         self, user_id, payload, client_id=None, lifespan=None, redirect_uri=None
@@ -751,7 +757,7 @@ class KeycloakAdmin:
             data=json.dumps(payload),
             **params_query
         )
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPutError)
 
     def send_verify_email(self, user_id, client_id=None, redirect_uri=None):
         """
@@ -769,7 +775,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_SEND_VERIFY_EMAIL.format(**params_path), data={}, **params_query
         )
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPutError)
 
     def get_sessions(self, user_id):
         """
@@ -854,7 +860,7 @@ class KeycloakAdmin:
         # went through the tree without hits
         return None
 
-    def get_group_members(self, group_id, **query):
+    def get_group_members(self, group_id, query=None):
         """
         Get members by group id. Returns group members
 
@@ -866,6 +872,7 @@ class KeycloakAdmin:
         (see https://www.keycloak.org/docs-api/18.0/rest-api/index.html#_getmembers)
         :return: Keycloak server response (UserRepresentation)
         """
+        query = query or {}
         params_path = {"realm-name": self.realm_name, "id": group_id}
         url = URL_ADMIN_GROUP_MEMBERS.format(**params_path)
 
@@ -914,7 +921,7 @@ class KeycloakAdmin:
         GroupRepresentation
         https://www.keycloak.org/docs-api/18.0/rest-api/#_grouprepresentation
 
-        :return: Http response
+        :return: Group id for newly created group or None for an existing group
         """
 
         if parent is None:
@@ -931,9 +938,14 @@ class KeycloakAdmin:
                 URL_ADMIN_GROUP_CHILD.format(**params_path), data=json.dumps(payload)
             )
 
-        return raise_error_from_response(
+        raise_error_from_response(
             data_raw, KeycloakPostError, expected_codes=[201], skip_exists=skip_exists
         )
+        try:
+            _last_slash_idx = data_raw.headers["Location"].rindex("/")
+            return data_raw.headers["Location"][_last_slash_idx + 1 :]  # noqa: E203
+        except KeyError:
+            return
 
     def update_group(self, group_id, payload):
         """
@@ -950,7 +962,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": group_id}
         data_raw = self.raw_put(URL_ADMIN_GROUP.format(**params_path), data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def group_set_permissions(self, group_id, enabled=True):
         """
@@ -966,7 +978,7 @@ class KeycloakAdmin:
             URL_ADMIN_GROUP_PERMISSIONS.format(**params_path),
             data=json.dumps({"enabled": enabled}),
         )
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPutError)
 
     def group_user_add(self, user_id, group_id):
         """
@@ -979,7 +991,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": user_id, "group-id": group_id}
         data_raw = self.raw_put(URL_ADMIN_USER_GROUP.format(**params_path), data=None)
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def group_user_remove(self, user_id, group_id):
         """
@@ -992,7 +1004,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": user_id, "group-id": group_id}
         data_raw = self.raw_delete(URL_ADMIN_USER_GROUP.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def delete_group(self, group_id):
         """
@@ -1004,7 +1016,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": group_id}
         data_raw = self.raw_delete(URL_ADMIN_GROUP.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_clients(self):
         """
@@ -1064,7 +1076,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": client_id}
         data_raw = self.raw_get(URL_ADMIN_CLIENT_AUTHZ_SETTINGS.format(**params_path))
-        return data_raw
+        return raise_error_from_response(data_raw, KeycloakGetError)
 
     def create_client_authz_resource(self, client_id, payload, skip_exists=False):
         """
@@ -1178,7 +1190,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": client_id}
         data_raw = self.raw_get(URL_ADMIN_CLIENT_AUTHZ_SCOPES.format(**params_path))
-        return data_raw
+        return raise_error_from_response(data_raw, KeycloakGetError)
 
     def get_client_authz_permissions(self, client_id):
         """
@@ -1191,7 +1203,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": client_id}
         data_raw = self.raw_get(URL_ADMIN_CLIENT_AUTHZ_PERMISSIONS.format(**params_path))
-        return data_raw
+        return raise_error_from_response(data_raw, KeycloakGetError)
 
     def get_client_authz_policies(self, client_id):
         """
@@ -1204,7 +1216,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": client_id}
         data_raw = self.raw_get(URL_ADMIN_CLIENT_AUTHZ_POLICIES.format(**params_path))
-        return data_raw
+        return raise_error_from_response(data_raw, KeycloakGetError)
 
     def get_client_service_account_user(self, client_id):
         """
@@ -1248,7 +1260,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": client_id}
         data_raw = self.raw_put(URL_ADMIN_CLIENT.format(**params_path), data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def delete_client(self, client_id):
         """
@@ -1263,7 +1275,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": client_id}
         data_raw = self.raw_delete(URL_ADMIN_CLIENT.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_client_installation_provider(self, client_id, provider_id):
         """
@@ -1410,7 +1422,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": client_role_id, "role-name": role_name}
         data_raw = self.raw_delete(URL_ADMIN_CLIENT_ROLE.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def assign_client_role(self, user_id, client_id, roles):
         """
@@ -1495,7 +1507,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_REALM_ROLES_ROLE_BY_NAME.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def delete_realm_role(self, role_name):
         """
@@ -1506,7 +1518,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "role-name": role_name}
         data_raw = self.raw_delete(URL_ADMIN_REALM_ROLES_ROLE_BY_NAME.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def add_composite_realm_roles_to_role(self, role_name, roles):
         """
@@ -1540,7 +1552,7 @@ class KeycloakAdmin:
             URL_ADMIN_REALM_ROLES_COMPOSITE_REALM_ROLE.format(**params_path),
             data=json.dumps(payload),
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_composite_realm_roles_of_role(self, role_name):
         """
@@ -1584,7 +1596,7 @@ class KeycloakAdmin:
         data_raw = self.raw_delete(
             URL_ADMIN_USER_REALM_ROLES.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_realm_roles_of_user(self, user_id):
         """
@@ -1648,7 +1660,7 @@ class KeycloakAdmin:
         data_raw = self.raw_delete(
             URL_ADMIN_GROUPS_REALM_ROLES.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_group_realm_roles(self, group_id):
         """
@@ -1706,7 +1718,7 @@ class KeycloakAdmin:
         data_raw = self.raw_delete(
             URL_ADMIN_GROUPS_CLIENT_ROLES.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_client_roles_of_user(self, user_id, client_id):
         """
@@ -1761,7 +1773,7 @@ class KeycloakAdmin:
         data_raw = self.raw_delete(
             URL_ADMIN_USER_CLIENT_ROLES.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_authentication_flows(self):
         """
@@ -1836,7 +1848,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": flow_id}
         data_raw = self.raw_delete(URL_ADMIN_FLOW.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_authentication_flow_executions(self, flow_alias):
         """
@@ -1865,7 +1877,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_FLOWS_EXECUTIONS.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[202, 204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[202, 204])
 
     def get_authentication_flow_execution(self, execution_id):
         """
@@ -1911,7 +1923,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": execution_id}
         data_raw = self.raw_delete(URL_ADMIN_FLOWS_EXECUTION.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def create_authentication_flow_subflow(self, payload, flow_alias, skip_exists=False):
         """
@@ -1960,7 +1972,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_AUTHENTICATOR_CONFIG.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def delete_authenticator_config(self, config_id):
         """
@@ -1973,8 +1985,7 @@ class KeycloakAdmin:
 
         params_path = {"realm-name": self.realm_name, "id": config_id}
         data_raw = self.raw_delete(URL_ADMIN_AUTHENTICATOR_CONFIG.format(**params_path))
-
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def sync_users(self, storage_id, action):
         """
@@ -2054,7 +2065,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_CLIENT_SCOPE.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def add_mapper_to_client_scope(self, client_scope_id, payload):
         """
@@ -2091,8 +2102,7 @@ class KeycloakAdmin:
         }
 
         data_raw = self.raw_delete(URL_ADMIN_CLIENT_SCOPES_MAPPERS.format(**params_path))
-
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def update_mapper_in_client_scope(self, client_scope_id, protocol_mapper_id, payload):
         """
@@ -2116,7 +2126,7 @@ class KeycloakAdmin:
             URL_ADMIN_CLIENT_SCOPES_MAPPERS.format(**params_path), data=json.dumps(payload)
         )
 
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def get_default_default_client_scopes(self):
         """
@@ -2137,7 +2147,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": scope_id}
         data_raw = self.raw_delete(URL_ADMIN_DEFAULT_DEFAULT_CLIENT_SCOPE.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def add_default_default_client_scope(self, scope_id):
         """
@@ -2151,7 +2161,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_DEFAULT_DEFAULT_CLIENT_SCOPE.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def get_default_optional_client_scopes(self):
         """
@@ -2172,7 +2182,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "id": scope_id}
         data_raw = self.raw_delete(URL_ADMIN_DEFAULT_OPTIONAL_CLIENT_SCOPE.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def add_default_optional_client_scope(self, scope_id):
         """
@@ -2186,7 +2196,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_DEFAULT_OPTIONAL_CLIENT_SCOPE.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def add_mapper_to_client(self, client_id, payload):
         """
@@ -2225,7 +2235,7 @@ class KeycloakAdmin:
             URL_ADMIN_CLIENT_PROTOCOL_MAPPER.format(**params_path), data=json.dumps(payload)
         )
 
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def remove_client_mapper(self, client_id, client_mapper_id):
         """
@@ -2243,8 +2253,7 @@ class KeycloakAdmin:
         }
 
         data_raw = self.raw_delete(URL_ADMIN_CLIENT_PROTOCOL_MAPPER.format(**params_path))
-
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def generate_client_secrets(self, client_id):
         """
@@ -2335,7 +2344,7 @@ class KeycloakAdmin:
         data_raw = self.raw_put(
             URL_ADMIN_COMPONENT.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def delete_component(self, component_id):
         """
@@ -2347,7 +2356,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name, "component-id": component_id}
         data_raw = self.raw_delete(URL_ADMIN_COMPONENT.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_keys(self):
         """
@@ -2386,7 +2395,7 @@ class KeycloakAdmin:
         """
         params_path = {"realm-name": self.realm_name}
         data_raw = self.raw_put(URL_ADMIN_EVENTS.format(**params_path), data=json.dumps(payload))
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
 
     def raw_get(self, *args, **kwargs):
         """
@@ -2531,7 +2540,7 @@ class KeycloakAdmin:
         data_raw = self.raw_delete(
             URL_ADMIN_DELETE_USER_ROLE.format(**params_path), data=json.dumps(payload)
         )
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
 
     def get_client_sessions_stats(self):
         """
